
//////////////////////////////////////////////////////
//                 GLOBAL VARIABLES                 //
//////////////////////////////////////////////////////

var dimension = 3;

var X = new Array(new Array(0, 100), new Array(0, 100), new Array(0, 80));

var epsilon = 0.0000000001;



//////////////////////////////////////////////////////
//                 LOCAL VARIABLES                 //
//////////////////////////////////////////////////////

var A = new Array(200);
var w = new Array(A.length);

var delta = 0.4;
var C = 0.4;

var random_input = false; // true = create random input data



//////////////////////////////////////////////////////
//                   CALCULATIONS                   //
//////////////////////////////////////////////////////

// initialize problem //

function init_problem () {
  if (random_input) for (var k = 0; k < A.length; k++) {
    var ax = 30+40*Math.random();
    var ay = 30+40*Math.random();
    var ar = 15+20*Math.random();
    for (var k = 0; k < A.length; k++) {
      var ap = Math.random()*Math.PI*2;
      A[k] = new Array(Math.floor(ax+ar*Math.cos(ap)), Math.floor(ay+ar*Math.sin(ap)));
      if (Math.random() > 0.015*ar) A[k] = new Array(Math.floor(Math.random()*100),Math.floor(Math.random()*100));
    }
  } else {
    A = new Array(
      new Array(81, 48), new Array(72, 18), new Array(65, 92), new Array(23, 90), new Array(66, 36),
      new Array(83, 52), new Array(25, 64), new Array(35, 42), new Array(85, 72), new Array(97, 39),
      new Array(47, 35), new Array(89, 13), new Array(24, 64), new Array(29, 50), new Array(25, 62),
      new Array(39, 89), new Array(9, 48), new Array(37, 27), new Array(34, 85), new Array(28, 75),
      new Array(6, 30), new Array(16, 78), new Array(85, 68), new Array(23, 11), new Array(39, 9),
      new Array(12, 48), new Array(84, 70), new Array(70, 45), new Array(74, 78), new Array(8, 30),
      new Array(74, 87), new Array(18, 32), new Array(85, 58), new Array(85, 67), new Array(72, 87),
      new Array(77, 85), new Array(27, 73), new Array(84, 73), new Array(60, 45), new Array(45, 74),
      new Array(67, 91), new Array(64, 92), new Array(36, 41), new Array(76, 86), new Array(41, 90),
      new Array(56, 94), new Array(65, 6), new Array(39, 39), new Array(34, 89), new Array(57, 34)
    );
  }
}

// objective function //

function distance (a1, a2, b1, b2) {
  return Math.sqrt( (a1-b1)*(a1-b1) + (a2-b2)*(a2-b2) );
}

function distance_r (a1, a2, b1, b2, r) {
  var h = distance(a1, a2, b1, b2) - r;
  return - h*h/delta;
}

function f (x) {
  var sum = 0;
  for (var k = 0; k < A.length; k++) sum -= Math.exp( distance_r(A[k][0],A[k][1],x[0],x[1],x[2]) );
  return sum + C*x[2];
}

// lower bounds //

var center = new Array(dimension);
var fac_g = 0;
var fac_dg0 = 0;
var fac_dg1 = 0;
var fac_dg2 = 0;

function minv (v) {
  var h = v[0];
  for (var i = 1; i < v.length; i++) if (v[i] < h) h = v[i];
  return h;
}

function g (x, k) {
  return 2*( (x[0]-A[k][0])*(x[0]-A[k][0]) + (x[1]-A[k][1])*(x[1]-A[k][1]) + x[2]*x[2]);
}

function h (x, k) {
  var t = distance(A[k][0],A[k][1],x[0],x[1]) + x[2];
  return t*t;
}

function calculate_hyperplane_fac (k) {
  fac_g = g(center, k);
  fac_dg0 = 4*(center[0]-A[k][0]);
  fac_dg1 = 4*(center[1]-A[k][1]);
  fac_dg2 = 4*center[2];
}

function hyperplane (x) {
  return fac_g + fac_dg0 * (x[0]-center[0]) + fac_dg1 * (x[1]-center[1]) + fac_dg2 * (x[2]-center[2]);
}

function LB_1 (I) {
  for (var i = 0; i < dimension; i++) center[i] = 0.5 * (I[i][0]+I[i][1]);
  var v = new Array(8);
  for (var i = 0; i < v.length; i++) v[i] = 0;

  var xv = new Array(dimension);

  for (var k = 0; k < A.length; k++) {
    calculate_hyperplane_fac(k);
    for (var i = 0; i < v.length; i++) {
      for (var j = 0; j < dimension; j++) xv[j] = ( (i >> j) & 1 == 1 ? I[j][0] : I[j][1]);
      v[i] -= Math.exp( (-hyperplane(xv) + h(xv,k))/delta  );
      if (k == 0) v[i] += C*xv[2];
    }
  }
 
  return minv(v);
}

function distance_I (I0, I1, k) {
  return IA.sqrt( IA.add( IA.sq(IA.add_scalar(-A[k][0],I0)), IA.sq(IA.add_scalar(-A[k][1],I1)) ) );
}

function distance_r_I (I0, I1, k, I2) {
  return IA.mult_scalar(-1/delta,IA.sq( IA.sub( distance_I(I0,I1,k), I2) ));
}

function LB_2 (I) {
  var H = new Array(0,0);
  for (var k = 0; k < A.length; k++) H = IA.sub(H, IA.exp(distance_r_I(I[0],I[1],k,I[2])) );
  return H[0] + C*I[2][0];
}

function LB (I) {
  var hLB1 = LB_1(I);
  var hLB2 = LB_2(I);
  return (hLB1 >= hLB2 ? hLB1 : hLB2);
}

 

//////////////////////////////////////////////////////
//              PROBLEM VISUALIZATION               //
//////////////////////////////////////////////////////

function plot_problem (y) {
  PV.set_size(0, 100, 0, 100, 1);

  PV.circle(y[0], y[1], y[2], '#DE102D');
  for (var k = 0; k < A.length; k++) PV.dot(A[k][0], A[k][1]);
  
  PV.text(13, 96, A.length + " demand points");
  PV.head("Problem Visualization");
}
